@@ -2,7 +2,7 @@
 #
 # Python Bindings for LZMA
 #
-# Copyright (c) 2004-2014 by Joachim Bauch, mail@joachim-bauch.de
+# Copyright (c) 2004-2015 by Joachim Bauch, mail@joachim-bauch.de
 # 7-Zip Copyright (C) 1999-2010 Igor Pavlov
 # LZMA SDK Copyright (C) 1999-2010 Igor Pavlov
 #
@@ -25,6 +25,7 @@
 """Read from and write to 7zip format archives.
 """
 
+from array import array
 from binascii import unhexlify
 from datetime import datetime
 import pylzma
@@ -86,6 +87,22 @@ except NameError:
     # Python 3.x
     long = int
 
+try:
+    xrange
+except NameError:
+    # Python 3.x
+    xrange = range
+
+IS_PYTHON3 = sys.version_info[0] == 3
+
+NEED_BYTESWAP = sys.byteorder != 'little'
+
+if array('L').itemsize == 4:
+    ARRAY_TYPE_UINT32 = 'L'
+else:
+    assert array('I').itemsize == 4
+    ARRAY_TYPE_UINT32 = 'I'
+
 READ_BLOCKSIZE                   = 16384
 
 MAGIC_7Z                         = unhexlify('377abcaf271c')  # '7z\xbc\xaf\x27\x1c'
@@ -132,6 +149,20 @@ def toTimestamp(filetime):
     # FILETIME is 100-nanosecond intervals since 1601/01/01 (UTC)
     return (filetime / 10000000.0) + TIMESTAMP_ADJUST
 
+def calculate_crc32(data, value=None, blocksize=1024*1024):
+    """Calculate CRC32 of strings with arbitrary lengths."""
+    length = len(data)
+    pos = blocksize
+    if value:
+        value = crc32(data[:pos], value)
+    else:
+        value = crc32(data[:pos])
+    while pos < length:
+        value = crc32(data[pos:pos+blocksize], value)
+        pos += blocksize
+
+    return value & 0xffffffff
+
 class ArchiveError(Exception):
     pass
 
@@ -153,6 +184,9 @@ class NoPasswordGivenError(DecryptionError):
 class WrongPasswordError(DecryptionError):
     pass
 
+class DecompressionError(ArchiveError):
+    pass
+
 class ArchiveTimestamp(long):
     """Windows FILETIME timestamp."""
     
@@ -174,9 +208,9 @@ class Base(object):
     def _read64Bit(self, file):
         b = ord(file.read(1))
         mask = 0x80
-        for i in range(8):
+        for i in xrange(8):
             if b & mask == 0:
-                bytes = list(unpack('%dB' % i, file.read(i)))
+                bytes = array('B', file.read(i))
                 bytes.reverse()
                 value = (bytes and reduce(lambda x, y: x << 8 | y, bytes)) or 0
                 highpart = b & (mask - 1)
@@ -193,7 +227,7 @@ class Base(object):
         result = []
         b = 0
         mask = 0
-        for i in range(count):
+        for i in xrange(count):
             if mask == 0:
                 b = ord(file.read(1))
                 mask = 0x80
@@ -203,7 +237,7 @@ class Base(object):
         return result
 
     def checkcrc(self, crc, data):
-        check = crc32(data) & 0xffffffff
+        check = calculate_crc32(data)
         return crc == check
 
 
@@ -215,11 +249,11 @@ class PackInfo(Base):
         self.numstreams = self._read64Bit(file)
         id = file.read(1)
         if id == PROPERTY_SIZE:
-            self.packsizes = [self._read64Bit(file) for x in range(self.numstreams)]
+            self.packsizes = [self._read64Bit(file) for x in xrange(self.numstreams)]
             id = file.read(1)
             
             if id == PROPERTY_CRC:
-                self.crcs = [self._read64Bit(file) for x in range(self.numstreams)]
+                self.crcs = [self._read64Bit(file) for x in xrange(self.numstreams)]
                 id = file.read(1)
             
         if id != PROPERTY_END:
@@ -234,7 +268,7 @@ class Folder(Base):
         self.digestdefined = False
         totalin = 0
         self.totalout = 0
-        for i in range(numcoders):
+        for i in xrange(numcoders):
             while True:
                 b = ord(file.read(1))
                 methodsize = b & 0xf
@@ -259,51 +293,52 @@ class Folder(Base):
         
         numbindpairs = self.totalout - 1
         self.bindpairs = []
-        for i in range(numbindpairs):
+        for i in xrange(numbindpairs):
             self.bindpairs.append((self._read64Bit(file), self._read64Bit(file), ))
         
         numpackedstreams = totalin - numbindpairs
         self.packed_indexes = []
         if numpackedstreams == 1:
-            for i in range(totalin):
+            for i in xrange(totalin):
                 if self.findInBindPair(i) < 0:
                     self.packed_indexes.append(i)
         elif numpackedstreams > 1:
-            for i in range(numpackedstreams):
+            for i in xrange(numpackedstreams):
                 self.packed_indexes.append(self._read64Bit(file))
 
     def getUnpackSize(self):
         if not self.unpacksizes:
             return 0
             
-        r = list(range(len(self.unpacksizes)))
-        r.reverse()
-        for i in r:
+        for i in xrange(len(self.unpacksizes)-1, -1, -1):
             if self.findOutBindPair(i):
                 return self.unpacksizes[i]
         
         raise TypeError('not found')
 
     def findInBindPair(self, index):
-        for idx in range(len(self.bindpairs)):
-            a, b = self.bindpairs[idx]
+        for idx, (a, b) in enumerate(self.bindpairs):
             if a == index:
                 return idx
         return -1
 
     def findOutBindPair(self, index):
-        for idx in range(len(self.bindpairs)):
-            a, b = self.bindpairs[idx]
+        for idx, (a, b) in enumerate(self.bindpairs):
             if b == index:
                 return idx
         return -1
-        
+
+    def isEncrypted(self):
+        return COMPRESSION_METHOD_7Z_AES256_SHA256 in [x['method'] for x in self.coders]
+
 class Digests(Base):
     """ holds a list of checksums """
     
     def __init__(self, file, count):
         self.defined = self._readBoolean(file, count, checkall=1)
-        self.crcs = [unpack('<L', file.read(4))[0] for x in range(count)]
+        self.crcs = array(ARRAY_TYPE_UINT32, file.read(4*count))
+        if NEED_BYTESWAP:
+            self.crcs.byteswap()
     
 UnpackDigests = Digests
 
@@ -318,7 +353,7 @@ class UnpackInfo(Base):
         self.folders = []
         external = file.read(1)
         if external == unhexlify('00'):
-            self.folders = [Folder(file) for x in range(self.numfolders)]
+            self.folders = [Folder(file) for x in xrange(self.numfolders)]
         elif external == unhexlify('01'):
             self.datastreamidx = self._read64Bit(file)
         else:
@@ -329,13 +364,12 @@ class UnpackInfo(Base):
             raise FormatError('coders unpack size id expected but %s found' % repr(id))
         
         for folder in self.folders:
-            folder.unpacksizes = [self._read64Bit(file) for x in range(folder.totalout)]
+            folder.unpacksizes = [self._read64Bit(file) for x in xrange(folder.totalout)]
             
         id = file.read(1)
         if id == PROPERTY_CRC:
             digests = UnpackDigests(file, self.numfolders)
-            for idx in range(self.numfolders):
-                folder = self.folders[idx]
+            for idx, folder in enumerate(self.folders):
                 folder.digestdefined = digests.defined[idx]
                 folder.crc = digests.crcs[idx]
                 
@@ -352,18 +386,16 @@ class SubstreamsInfo(Base):
         self.digestsdefined = []
         id = file.read(1)
         if id == PROPERTY_NUM_UNPACK_STREAM:
-            self.numunpackstreams = [self._read64Bit(file) for x in range(numfolders)]
+            self.numunpackstreams = [self._read64Bit(file) for x in xrange(numfolders)]
             id = file.read(1)
-        else:        
-            self.numunpackstreams = []
-            for idx in range(numfolders):
-                self.numunpackstreams.append(1)
+        else:
+            self.numunpackstreams = [1]*numfolders
         
         if id == PROPERTY_SIZE:
-            sum = 0
             self.unpacksizes = []
-            for i in range(len(self.numunpackstreams)):
-                for j in range(1, self.numunpackstreams[i]):
+            for i in xrange(len(self.numunpackstreams)):
+                sum = 0
+                for j in xrange(1, self.numunpackstreams[i]):
                     size = self._read64Bit(file)
                     self.unpacksizes.append(size)
                     sum += size
@@ -373,7 +405,7 @@ class SubstreamsInfo(Base):
 
         numdigests = 0
         numdigeststotal = 0
-        for i in range(numfolders):
+        for i in xrange(numfolders):
             numsubstreams = self.numunpackstreams[i]
             if numsubstreams != 1 or not folders[i].digestdefined:
                 numdigests += numsubstreams
@@ -382,14 +414,14 @@ class SubstreamsInfo(Base):
         if id == PROPERTY_CRC:
             digests = Digests(file, numdigests)
             didx = 0
-            for i in range(numfolders):
+            for i in xrange(numfolders):
                 folder = folders[i]
                 numsubstreams = self.numunpackstreams[i]
                 if numsubstreams == 1 and folder.digestdefined:
                     self.digestsdefined.append(True)
                     self.digests.append(folder.crc)
                 else:
-                    for j in range(numsubstreams):
+                    for j in xrange(numsubstreams):
                         self.digestsdefined.append(digests.defined[didx])
                         self.digests.append(digests.crcs[didx])
                         didx += 1
@@ -431,7 +463,7 @@ class FilesInfo(Base):
         
         # NOTE: the "external" flag is currently ignored, should be 0x00
         external = file.read(1)
-        for i in range(len(files)):
+        for i in xrange(len(files)):
             if defined[i]:
                 files[i][name] = ArchiveTimestamp(self._readReal64Bit(file)[0])
             else:
@@ -439,7 +471,7 @@ class FilesInfo(Base):
 
     def __init__(self, file):
         self.numfiles = self._read64Bit(file)
-        self.files = [{'emptystream': False} for x in range(self.numfiles)]
+        self.files = [{'emptystream': False} for x in xrange(self.numfiles)]
         numemptystreams = 0
         while True:
             typ = self._read64Bit(file)
@@ -486,9 +518,8 @@ class FilesInfo(Base):
                 self._readTimes(buffer, self.files, 'lastwritetime')
             elif typ == PROPERTY_ATTRIBUTES:
                 defined = self._readBoolean(buffer, self.numfiles, checkall=1)
-                for i in range(self.numfiles):
-                    f = self.files[i]
-                    if defined[i]:
+                for idx, f in enumerate(self.files):
+                    if defined[idx]:
                         f['attributes'] = unpack('<L', buffer.read(4))[0]
                     else:
                         f['attributes'] = None
@@ -522,18 +553,18 @@ class Header(Base):
 class ArchiveFile(Base):
     """ wrapper around a file in the archive """
     
-    def __init__(self, info, start, src_start, size, folder, archive, maxsize=None):
+    def __init__(self, info, start, src_start, folder, archive, maxsize=None):
         self.digest = None
         self._archive = archive
         self._file = archive._file
         self._start = start
         self._src_start = src_start
         self._folder = folder
-        self.size = size
         # maxsize is only valid for solid archives
         self._maxsize = maxsize
         for k, v in info.items():
             setattr(self, k, v)
+        self.size = self.uncompressed = self._uncompressed[-1]
         if not hasattr(self, 'filename'):
             # compressed file is stored without a name, generate one
             try:
@@ -553,7 +584,7 @@ class ArchiveFile(Base):
         }
 
     def _is_encrypted(self):
-        return COMPRESSION_METHOD_7Z_AES256_SHA256 in [x['method'] for x in self._folder.coders]
+        return self._folder.isEncrypted()
 
     def reset(self):
         self.pos = 0
@@ -563,6 +594,7 @@ class ArchiveFile(Base):
             raise TypeError("file has no coder informations")
         
         data = None
+        level = 0
         for coder in self._folder.coders:
             method = coder['method']
             decoder = None
@@ -573,17 +605,20 @@ class ArchiveFile(Base):
             if decoder is None:
                 raise UnsupportedCompressionMethodError(repr(coder['method']))
             
-            data = getattr(self, decoder)(coder, data)
+            data = getattr(self, decoder)(coder, data, level)
+            level += 1
         
         return data
     
-    def _read_copy(self, coder, input):
+    def _read_copy(self, coder, input, level):
+        size = self._uncompressed[level]
         if not input:
             self._file.seek(self._src_start)
-            input = self._file.read(self.uncompressed)
-        return input[self._start:self._start+self.size]
+            input = self._file.read(size)
+        return input[self._start:self._start+size]
     
-    def _read_from_decompressor(self, coder, decompressor, input, checkremaining=False, with_cache=False):
+    def _read_from_decompressor(self, coder, decompressor, input, level, checkremaining=False, with_cache=False):
+        size = self._uncompressed[level]
         data = ''
         idx = 0
         cnt = 0
@@ -592,7 +627,7 @@ class ArchiveFile(Base):
             decompressor.decompress(properties)
         total = self.compressed
         if not input and total is None:
-            remaining = self._start+self.size
+            remaining = self._start+size
             out = BytesIO()
             cache = getattr(self._folder, '_decompress_cache', None)
             if cache is not None:
@@ -609,7 +644,8 @@ class ArchiveFile(Base):
                     tmp = decompressor.decompress(data, remaining)
                 else:
                     tmp = decompressor.decompress(data)
-                assert len(tmp) > 0
+                if not tmp and not data:
+                    raise DecompressionError('end of stream while decompressing')
                 out.write(tmp)
                 remaining -= len(tmp)
             
@@ -623,41 +659,46 @@ class ArchiveFile(Base):
                 self._file.seek(self._src_start)
                 input = self._file.read(total)
             if checkremaining:
-                data = decompressor.decompress(input, self._start+self.size)
+                data = decompressor.decompress(input, self._start+size)
             else:
                 data = decompressor.decompress(input)
-        return data[self._start:self._start+self.size]
+        return data[self._start:self._start+size]
     
-    def _read_lzma(self, coder, input):
-        dec = pylzma.decompressobj(maxlength=self._start+self.size)
+    def _read_lzma(self, coder, input, level):
+        size = self._uncompressed[level]
+        dec = pylzma.decompressobj(maxlength=self._start+size)
         try:
-            return self._read_from_decompressor(coder, dec, input, checkremaining=True, with_cache=True)
+            return self._read_from_decompressor(coder, dec, input, level, checkremaining=True, with_cache=True)
         except ValueError:
             if self._is_encrypted():
                 raise WrongPasswordError('invalid password')
             
             raise
         
-    def _read_zip(self, coder, input):
+    def _read_zip(self, coder, input, level):
         dec = zlib.decompressobj(-15)
-        return self._read_from_decompressor(coder, dec, input, checkremaining=True)
+        return self._read_from_decompressor(coder, dec, input, level, checkremaining=True)
         
-    def _read_bzip(self, coder, input):
+    def _read_bzip(self, coder, input, level):
         dec = bz2.BZ2Decompressor()
-        return self._read_from_decompressor(coder, dec, input)
+        return self._read_from_decompressor(coder, dec, input, level)
     
-    def _read_7z_aes256_sha256(self, coder, input):
+    def _read_7z_aes256_sha256(self, coder, input, level):
         if not self._archive.password:
             raise NoPasswordGivenError()
         
         # TODO: this needs some sanity checks
-        firstbyte = ord(coder['properties'][0])
+        firstbyte = coder['properties'][0]
+        if not IS_PYTHON3:
+            firstbyte = ord(firstbyte)
         numcyclespower = firstbyte & 0x3f
         if firstbyte & 0xc0 != 0:
             saltsize = (firstbyte >> 7) & 1
             ivsize = (firstbyte >> 6) & 1
             
-            secondbyte = ord(coder['properties'][1])
+            secondbyte = coder['properties'][1]
+            if not IS_PYTHON3:
+                secondbyte = ord(secondbyte)
             saltsize += (secondbyte >> 4)
             ivsize += (secondbyte & 0x0f)
             
@@ -668,16 +709,16 @@ class ArchiveFile(Base):
             assert len(iv) == ivsize
             assert numcyclespower <= 24
             if ivsize < 16:
-                iv += '\x00'*(16-ivsize)
+                iv += bytes('\x00'*(16-ivsize), 'ascii')
         else:
-            salt = iv = ''
+            salt = iv = bytes('', 'ascii')
         
         password = self._archive.password.encode('utf-16-le')
         key = pylzma.calculate_key(password, numcyclespower, salt=salt)
         cipher = pylzma.AESDecrypt(key, iv=iv)
         if not input:
             self._file.seek(self._src_start)
-            uncompressed_size = self.uncompressed
+            uncompressed_size = self._uncompressed[level]
             if uncompressed_size & 0x0f:
                 # we need a multiple of 16 bytes
                 uncompressed_size += 16 - (uncompressed_size & 0x0f)
@@ -707,12 +748,12 @@ class Archive7z(Base):
 
         self.startheadercrc = unpack('<L', file.read(4))[0]
         self.nextheaderofs, data = self._readReal64Bit(file)
-        crc = crc32(data)
+        crc = calculate_crc32(data)
         self.nextheadersize, data = self._readReal64Bit(file)
-        crc = crc32(data, crc)
+        crc = calculate_crc32(data, crc)
         data = file.read(4)
         self.nextheadercrc = unpack('<L', data)[0]
-        crc = crc32(data, crc) & 0xffffffff
+        crc = calculate_crc32(data, crc)
         if crc != self.startheadercrc:
             raise FormatError('invalid header data')
         self.afterheader = file.tell()
@@ -733,18 +774,29 @@ class Archive7z(Base):
             streams = StreamsInfo(buffer)
             file.seek(self.afterheader + 0)
             data = bytes('', 'ascii')
-            idx = 0
+            src_start = self.afterheader
             for folder in streams.unpackinfo.folders:
-                file.seek(streams.packinfo.packpos, 1)
-                props = folder.coders[0]['properties']
-                for idx in range(len(streams.packinfo.packsizes)):
-                    tmp = file.read(streams.packinfo.packsizes[idx])
-                    data += pylzma.decompress(props+tmp, maxlength=folder.unpacksizes[idx])
-                
+                if folder.isEncrypted() and not password:
+                    raise NoPasswordGivenError()
+
+                src_start += streams.packinfo.packpos
+                uncompressed = folder.unpacksizes
+                if not isinstance(uncompressed, (list, tuple)):
+                    uncompressed = [uncompressed] * len(folder.coders)
+                info = {
+                    'compressed': streams.packinfo.packsizes[0],
+                    '_uncompressed': uncompressed,
+                }
+                tmp = ArchiveFile(info, 0, src_start, folder, self)
+                folderdata = tmp.read()
+                src_start += uncompressed[-1]
+
                 if folder.digestdefined:
-                    if not self.checkcrc(folder.crc, data):
+                    if not self.checkcrc(folder.crc, folderdata):
                         raise FormatError('invalid block data')
-                        
+
+                data += folderdata
+
             buffer = BytesIO(data)
         
         self.files = []
@@ -766,34 +818,38 @@ class Archive7z(Base):
         if hasattr(subinfo, 'unpacksizes'):
             unpacksizes = subinfo.unpacksizes
         else:
-            unpacksizes = [x.unpacksizes[0] for x in folders]
+            unpacksizes = [x.unpacksizes for x in folders]
         
         fidx = 0
         obidx = 0
+        streamidx = 0
         src_pos = self.afterheader
         pos = 0
-        folder_start = 0
         folder_pos = src_pos
         maxsize = (self.solid and packinfo.packsizes[0]) or None
-        for idx in range(files.numfiles):
-            info = files.files[idx]
+        for info in files.files:
             if info['emptystream']:
                 continue
             
             folder = folders[fidx]
-            folder.solid = subinfo.numunpackstreams[fidx] > 1
+            if streamidx == 0:
+                folder.solid = subinfo.numunpackstreams[fidx] > 1
+
             maxsize = (folder.solid and packinfo.packsizes[fidx]) or None
-            if folder.solid:
+            uncompressed = unpacksizes[obidx]
+            if not isinstance(uncompressed, (list, tuple)):
+                uncompressed = [uncompressed] * len(folder.coders)
+            if pos > 0:
                 # file is part of solid archive
                 info['compressed'] = None
-            elif obidx < len(packsizes):
+            elif fidx < len(packsizes):
                 # file is compressed
-                info['compressed'] = packsizes[obidx]
+                info['compressed'] = packsizes[fidx]
             else:
                 # file is not compressed
-                info['compressed'] = unpacksizes[obidx]
-            info['uncompressed'] = unpacksizes[obidx]
-            file = ArchiveFile(info, pos, src_pos, unpacksizes[obidx], folder, self, maxsize=maxsize)
+                info['compressed'] = uncompressed
+            info['_uncompressed'] = uncompressed
+            file = ArchiveFile(info, pos, src_pos, folder, self, maxsize=maxsize)
             if subinfo.digestsdefined[obidx]:
                 file.digest = subinfo.digests[obidx]
             self.files.append(file)
@@ -802,11 +858,13 @@ class Archive7z(Base):
             else:
                 src_pos += info['compressed']
             obidx += 1
-            if idx >= subinfo.numunpackstreams[fidx]+folder_start:
+            streamidx += 1
+            if streamidx >= subinfo.numunpackstreams[fidx]:
+                pos = 0
                 folder_pos += packinfo.packsizes[fidx]
                 src_pos = folder_pos
-                folder_start = idx
                 fidx += 1
+                streamidx = 0
         
         self.numfiles = len(self.files)
         self.filenames = list(map(lambda x: x.filename, self.files))
