@@ -1,7 +1,7 @@
 /*
  * Python Bindings for LZMA
  *
- * Copyright (c) 2004-2014 by Joachim Bauch, mail@joachim-bauch.de
+ * Copyright (c) 2004-2015 by Joachim Bauch, mail@joachim-bauch.de
  * 7-Zip Copyright (C) 1999-2010 Igor Pavlov
  * LZMA SDK Copyright (C) 1999-2010 Igor Pavlov
  *
@@ -96,6 +96,14 @@ PyInterpreterState* _pylzma_interpreterState;
     0,
 #endif
 
+#if defined(PY_SSIZE_T_CLEAN) && (PY_VERSION_HEX >= 0x02040000)
+#define PARSE_LENGTH_TYPE   Py_ssize_t
+#define PARSE_LENGTH_FORMAT "%zd"
+#else
+#define PARSE_LENGTH_TYPE   int
+#define PARSE_LENGTH_FORMAT "%d"
+#endif
+
 extern ISzAlloc allocator;
 
 #endif
