@@ -1,7 +1,7 @@
 /*
  * Python Bindings for LZMA
  *
- * Copyright (c) 2004-2014 by Joachim Bauch, mail@joachim-bauch.de
+ * Copyright (c) 2004-2015 by Joachim Bauch, mail@joachim-bauch.de
  * 7-Zip Copyright (C) 1999-2010 Igor Pavlov
  * LZMA SDK Copyright (C) 1999-2010 Igor Pavlov
  *
@@ -58,7 +58,7 @@ pylzma_compress(PyObject *self, PyObject *args, PyObject *kwargs)
     char *matchfinder = NULL;    // matchfinder algorithm
     int algorithm = 2;
     char *data;
-    int length;
+    PARSE_LENGTH_TYPE length;
 
     if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s#|iiiiiiiis", kwlist, &data, &length, &dictionary, &fastBytes,
                                                                   &literalContextBits, &literalPosBits, &posBits, &algorithm, &eos, &multithreading, &matchfinder))
